import basicSearch from 'en-select/utils/basic-search';

import KeyboardEventHandler from 'en-select/mixins/keyboard-event-handler';
import DocumentEvent from 'en-select/mixins/document-event';
import isOutsideClick from 'en-select/utils/is-outside-click';

const {
  Mixin,
  assert,
  set,
  get,
  getProperties,
  setProperties,
  getWithDefault,
  computed,
  isEmpty,
  on,
  Logger: { log, warn },
  run,
  $,
  A: EmberArray,
  String: { htmlSafe },
  RSVP: { Promise }
} = Em

const fallback = (fallback) => {
  return computed({
    get () {
      return fallback
    },
    set (key, value) {
      if (typeof value === 'undefined') {
        return fallback
      } else {
        return value
      }
    }
  })
}

export default Mixin.create(DocumentEvent, KeyboardEventHandler, {
  attributeBindings: ['tabindex', 'role'],
  documentEvents: ['click', 'resize', 'scroll'],

  tabindex: 0,
  role: 'listbox',

  canRemoveSelection: true,

  optionLabelPath: 'label',
  optionValuePath: 'id',

  isOpen: false,
  isDisabled: false,
  isClosing: false,
  isFixed: false,
  isInvalid: false,

  alignFromBottom: false,

  options: fallback(null),
  selections: fallback(null),
  focused: null,

  search: null,
  disableSearch: false,

  label: null,

  multiple: false,

  onDocumentClick (e) {
    const { isOpen, multiple, selectableCount } = getProperties(
      this, 'isOpen', 'multiple', 'selectableCount'
    )

    if (!isOpen) {
      return false
    }

    /*
     * If it's an outside click, we close in all cases.
    */

    if (isOutsideClick($(this.element), $(e.target))) {
      this.send('closeSelect')
    }

    /*
     * If it's an inside click (e.g. an option), we see
     * if it's a multiple select and then decide
    */

    if (multiple && selectableCount > 0) {
      return
    }

    this.send('closeSelect')
  },

  onDocumentScroll (e) {
    this._repositionDropdown()
  },

  onDocumentResize (e) {
    this._repositionDropdown()
  },

  _repositionDropdown () {
    /**
     * Unless the `isFixed` flag is passed in,
     * it will not attempt to reposition.
    */

    let isFixed = get(this, 'isFixed')
    if (!isFixed) return

    let dropdown = this.$('.en-select-options')
    let label    = this.$('.en-select-label')

    let offset   = $(this.element).offset()

    if (!offset) {
      warn('Could not find offset')
      return
    }

    let { top, left } = offset

    if (!top || !left) {
      warn('Could not find offset')
      return
    }

    /*
     * scrollTop is the value of how much the
     * user has scrolled from the top.
    */

    let scrollTop = $(window).scrollTop()

    /*
     * The offset of the dropdown is the offset of
     * the element (i.e. select) from the top minus
     * the amount the user has scrolled.
     *
     * For example, if the dropdown is 30px off the
     * top and the user has scrolled 1200px, the dropdown
     * should not be visible. It will have an offset of
     * -1170px.
    */

    let elOffset  = parseInt(top, 10) - parseInt(scrollTop, 10)

    let labelHeight = label.height()
    let labelWidth  = label.width()

    /*
     * Just adds the height of the label so the options don't
     * overlay the label.
     */

    let offsetTop = parseInt(elOffset, 10) + parseInt(labelHeight, 10) + 5

    let style = htmlSafe(`
      top: ${offsetTop}px;
      left: ${left}px;
      width: ${labelWidth}px;
    `)

    set(this, 'dropdownStyle', style)
  },

  selectionComponent: computed('multiple', function () {
    if (get(this, 'multiple')) {
      return 'en-multi-select-selection'
    } else {
      return 'en-select-selection'
    }
  }),

  hasSelections: computed.gt('selections.length', 0),

  /**
   *  Selection
   *
   *  Returns the first object from selections unless
   *  in multiple mode
   *
   * @return {Em.Object}
   */

  selection: computed('selections.[]', function () {
    const multiple   = get(this, 'multiple')
    const selections = get(this, 'selections')

    assert('[en-select] Cannot get selection in multiple mode.', !multiple)

    return selections.objectAt(0)
  }),

  hasSelection: computed.bool('selection'),

  /**
   * sortedOptions
   *
   * By default, it's sorted by `id`.
   *
   * @return {Em.Array}
   */

  sortBy: ['id:asc'],

  sortedOptions: computed('options.[]', 'sortBy.[]', function () {
    let options = get(this, 'options')
    let sortBy = get(this, 'sortBy')

    assert('[en-select] sortBy must be an array', sortBy instanceof Array)

    return EmberArray(options).sortBy(...sortBy)
  }),

  /**
   * searchResults
   *
   * Filters the options if the searched keyword is present in their `labelPath`.
   *
   * @return {Em.Array}
   */

  searchResults: computed('search', 'sortedOptions.[]', function () {
    const options   = get(this, 'sortedOptions')
    const search    = get(this, 'search')
    const labelPath = getWithDefault(this, 'optionLabelPath', 'label')

    return basicSearch(search, options, { labelPath })
  }),

  /**
   * @public
   *
   * Selectable options; the difference between the
   * `searchResults` array and `selections` array.
   *
   * @return {Em.Array}
   */

  selectable: computed.setDiff('searchResults', 'selections'),

  selectableCount: computed.readOnly('selectable.length'),

  keyDown (e) {
    let originalEvent = e.originalEvent
    let keyCode = e.which || e.keyCode

    let relevantKeyCodes = [38, 40, 13, 27]

    /*
      * For key events that we want to capture, we stop
      * their default behavior.
    */

    if (relevantKeyCodes.includes(keyCode)) {
      e.preventDefault()
      e.stopPropagation()
    }

    switch (keyCode) {
      case 38:
        this._move('up', originalEvent)
        break;

      case 40:
        if (get(this, 'isOpen')) {
          this._move('down', originalEvent)
        } else {
          this.send('openSelect')
        }
        break;

      case 13:
        if (get(this, 'isOpen')) {
          this._selectFocusedOption(originalEvent)
        } else {
          this.send('openSelect')
        }

        break;

      case 27:
        this.send('closeSelect')
        break;

      case 8:
        // When a user presses backspace, the select is multiple
        // and there is no search, we delete the last item

        if (get(this, 'multiple') && isEmpty(get(this, 'search'))) {
          this._removeLastSelection()
        }

        break;
    }
  },

  _selectFocusedOption (e) {
    const focused = get(this, 'focused')

    if (!isEmpty(focused)) {
      this.send('select', focused)
    }
  },

  _focusOnSearch () {
    run.scheduleOnce('afterRender', () => {
      const searchBox = this.$('input[type="text"]')
      if (searchBox) searchBox.focus()
    })
  },

  _select (value) {
    if (get(this, 'isDestroyed')) return

    let options    = get(this, 'options')
    let selections = get(this, 'selections')
    let multiple   = get(this, 'multiple')
    let valuePath  = get(this, 'optionValuePath')

    if (isEmpty(options)) {
      warn('[en-select] No options to select from')
      return
    }

    if (isEmpty(value)) {
      warn('[en-select] No value to select')
      return
    }

    let id

    /*
     * Figure out the ID of the object. If the value
     * we get is a string, we just match the IDs to
     * be sure.
     *
     * In case it's an object, we find the ID using the
     * `optionValuePath` and use that to find the selected
     * option.
    */

    if (typeof value === "object") {
      id = get(value, valuePath)
    } else {
      id = value
    }

    let filtered = EmberArray(options.filter(option => get(option, valuePath) === id))

    /*
     * In case it's a multiple select, we can just push to
     * the selections array.
     *
     * Otherwise, we just set the selections again.
    */

    if (multiple) {
      get(this, 'selections').addObjects(filtered)
    } else {
      set(this, 'selections', filtered)
    }

    return get(this, 'selections')
  },

  /**
   * @private
   * 
   * Removes elements from selections array
   *
   * @return {undefined}
   */

  _removeFromSelections (value) {
    if (isEmpty(value)) {
      return
    } else {
      get(this, 'selections').removeObject(value)
    }
  },

  _removeLastSelection () {
    let selections = this.get('selections')

    this._removeFromSelections(
      selections.objectAt(selections.get('length') - 1)
    )

    this.send('openSelect')
  },

  /**
   * @private
   * Updates selections whenever new attrs are received
   *
   */

  _selectFromValues () {
    let values = get(this, 'value')

    if (!values) return

    if (values instanceof Promise) {
      values.then(v => this._selectValues(v))
    } else {
      this._selectValues(values)
    }
  },

  _selectValues (values) {
    let multiple = get(this, 'multiple')

    if (multiple) {
      values.forEach(value => this._select(value))

    } else {
      this._select(values)
    }
  },

  init () {
    this._super(...arguments)
        
    if (get(this, 'options') === null) set(this, 'options', EmberArray([]))
    if (get(this, 'selections') === null) set(this, 'selections', EmberArray([]))
    
    this._selectFromValues()
  },

  didReceiveAttrs () {
    this._selectFromValues()
  },

  actions: {
    openSelect () {
      if (get(this, 'isDisabled')) {
        return
      }

      run(() => {
        set(this, 'isOpen', true)

        this._subscribeToEvents()
        this._repositionDropdown()
        this._focusOnSearch()
      })
    },

    closeSelect () {
      set(this, 'isClosing', true)

      run(() => {
        if (get(this, 'isDestroyed')) return

        set(this, 'isClosing', false)
        set(this, 'isOpen', false)

        this._teardown()

        this.element.focus()
      })
    },

    searchFor (keyword) {
      set(this, 'search', keyword)
      set(this, 'focused', get(this, 'searchResults').objectAt(0))

      if (get(this, 'isOpen')) {
        this.send('openSelect')
      }
    },

    /**
     * Focuses on a given option, and, if the event
     * is a keyboard event, it asks the scroller to
     * scroll to that option.
     */

    focus (option, event) {
      set(this, 'focused', option)

      if (event && (event instanceof KeyboardEvent)) {
        this.send('_scrollIfFocusNotVisible')
      }

      if (this.attrs['on-focus']) {
        this.attrs['on-focus'](option)
      }
    },

    select (option) {
      let multiple = get(this, 'multiple')

      this._select(option)
      this.attrs['on-select'](option)

      if (multiple) {
        run(() => this.$('.en-select-search input').focus())
      } else {
        this.send('closeSelect')
      }

      set(this, 'search', null)
      set(this, 'focused', null)
    },

    removeSelection (option = null) {
      const multiple = get(this, 'multiple')

      if (multiple) {
        this._removeFromSelections(option)
      } else {
        set(this, 'selections', Em.A([]))
      }

      if (this.attrs['on-remove']) {
        this.attrs['on-remove'](option)
      }
    },

    /**
     *  Scroll to the focused option.
     *
     *  Checks if the focused option is out of view and
     *  scrolls to it, if so.
     */

    _scrollIfFocusNotVisible () {
      const list = this.$('.en-select-options').get(0)
      const focused = this.$('.en-select-option.is-focused').get(0)

      if (!list || !focused) return

      /**
       * Calculate the focused element's distance from the options
       * list's start
       */

      let scrollTop = focused.offsetTop - list.offsetTop

      /**
       * Add its offsetHeight to calculate the offset from top including
       * the focused element
       */

      let scrollBottom = scrollTop + focused.offsetHeight

      /**
       * Now, if the focused item's offset from the top is more than the
       * list's height + the list's current scroll position from the top,
       * you want to scroll again.
       */

      if (scrollBottom > (list.offsetHeight + list.scrollTop - 100)) {
        list.scrollTop = scrollBottom - (list.offsetHeight - 100)
      } else if (scrollTop < list.scrollTop) {
        list.scrollTop = scrollTop
      }
    }
  }
})
